#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pathlib

import pytest

from bioio_base.io import pathlike_to_fs


@pytest.mark.parametrize(
    "enforce_exists",
    [True, False],
)
def test_pathlike_to_fs(enforce_exists: bool, sample_text_file: pathlib.Path) -> None:
    pathlike_to_fs(sample_text_file, enforce_exists, fs_kwargs=dict(anon=True))


@pytest.mark.parametrize(
    "enforce_exists",
    [
        pytest.param(
            True,
            marks=pytest.mark.xfail(raises=FileNotFoundError),
        ),
        (False),
    ],
)
def test_pathlib_to_fs_with_missing_file(
    enforce_exists: bool, tmp_path: pathlib.Path
) -> None:
    uri = tmp_path / "missing-file.txt"

    pathlike_to_fs(uri, enforce_exists, fs_kwargs=dict(anon=True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing

import pytest

from bioio_base.dimensions import Dimensions


def test_dimensions_getitem() -> None:
    dims = Dimensions("TCZYX", (1, 4, 75, 624, 924))
    assert dims["T"] == (1,)
    assert dims["T", "C"] == (1, 4)

    # out of order indexing
    assert dims["C", "T", "Y"] == (4, 1, 624)

    with pytest.raises(IndexError):
        dims["blarg"]
    with pytest.raises(IndexError):
        dims["blarg", "nope"]
    with pytest.raises(TypeError):
        # Ironic we have to type ignore this because uhhh
        # we are testing a TypeError
        dims[0]  # type: ignore

    assert dims.T == 1
    assert dims.order == "TCZYX"


@pytest.mark.parametrize(
    "dims, shape",
    [
        (["Z", "Y", "X"], (70, 980, 980)),
        pytest.param(
            "ZYXS",
            (70, 980, 980),
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            "YX",
            (70, 980, 980),
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_dimensions_mismatched_dims_len_and_shape_size(
    dims: typing.Collection[str],
    shape: typing.Tuple[int, ...],
) -> None:
    # Just check success
    assert Dimensions(dims, shape)


@pytest.mark.parametrize(
    "dims, shape",
    [
        (["Z", "Y", "X"], (70, 980, 980)),
        pytest.param(
            ["C", "ZY", "X"],
            (70, 980, 980),
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            ["YX"],
            (70, 980, 980),
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_dimensions_bad_iterable_of_characters(
    dims: typing.Collection[str],
    shape: typing.Tuple[int, ...],
) -> None:
    # Just check success
    assert Dimensions(dims, shape)



import typing

import dask.array as da
import fsspec
import numpy as np
import xarray as xr

from .dimensions import DEFAULT_DIMENSION_ORDER_LIST
from .reader import Reader


class NoopReader(Reader):
    """
    No-op (no operation) reader intended to be used in tests as a way to test utilities
    that utilize readers but are not trying to test any specific reader.

    NOT intended to be inherited by plug-in readers see ImageContainer instead.
    """

    _NUM_SCENES = 3
    _mock_data = np.arange(
        np.prod((_NUM_SCENES, 4, 5, 6, 7, 8)), dtype="uint16"
    ).reshape((_NUM_SCENES, 4, 5, 6, 7, 8))

    @staticmethod
    def _is_supported_image(
        fs: fsspec.AbstractFileSystem,
        path: str,
        **kwargs: typing.Any,
    ) -> bool:
        return True

    @property
    def scenes(self) -> typing.Tuple[str, ...]:
        return tuple(f"Image:{idx}" for idx in range(self._NUM_SCENES))

    def _read_delayed(self) -> xr.DataArray:
        return xr.DataArray(
            data=da.from_array(self._mock_data[self.current_scene_index]),
            dims=DEFAULT_DIMENSION_ORDER_LIST,
        )

    def _read_immediate(self) -> xr.DataArray:
        return xr.DataArray(
            data=self._mock_data[self.current_scene_index],
            dims=DEFAULT_DIMENSION_ORDER_LIST,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import dask.array as da
import numpy as np
import xarray as xr
from fsspec.spec import AbstractFileSystem
from ome_types import OME

from . import constants, exceptions, transforms, types
from .dimensions import DEFAULT_DIMENSION_ORDER, DimensionNames, Dimensions
from .image_container import ImageContainer
from .io import pathlike_to_fs
from .types import PhysicalPixelSizes

###############################################################################


class Reader(ImageContainer, ABC):
    """
    A small class to build standardized image reader objects that deal with the raw
    image and metadata.

    Parameters
    ----------
    image: Any
        Some type of object to read and follow the Reader specification.
    fs_kwargs: Dict[str, Any]
        Any specific keyword arguments to pass down to the fsspec created filesystem.
        Default: {}

    Notes
    -----
    It is up to the implementer of the Reader to decide which types they would like to
    accept (certain readers may not support buffers for example).
    """

    _xarray_dask_data: Optional[xr.DataArray] = None
    _xarray_data: Optional[xr.DataArray] = None
    _mosaic_xarray_dask_data: Optional[xr.DataArray] = None
    _mosaic_xarray_data: Optional[xr.DataArray] = None
    _dims: Optional[Dimensions] = None
    _metadata: Optional[Any] = None
    _scenes: Optional[Tuple[str, ...]] = None
    _current_scene_index: int = 0
    _current_resolution_level: int = 0
    # Do not default because they aren't used by all readers
    _fs: AbstractFileSystem
    _path: str

    @staticmethod
    @abstractmethod
    def _is_supported_image(
        fs: AbstractFileSystem,
        path: str,
        **kwargs: Any,
    ) -> bool:
        """
        The per-Reader implementation used to validate that an image is supported or not
        by the Reader itself.

        Parameters
        ----------
        fs: AbstractFileSystem
            The file system to used for reading.
        path: str
            The path to the file to read.
        kwargs: Any
            Any kwargs used for reading and validation of the file.

        Returns
        -------
        supported: bool
            Boolean value indicating if the file is supported by the reader.
        """

    @classmethod
    def is_supported_image(
        cls,
        image: types.ImageLike,
        fs_kwargs: Dict[str, Any] = {},
        **kwargs: Any,
    ) -> bool:
        """
        Asserts that the provided image like object is supported by the current Reader.

        Parameters
        ----------
        image: types.ImageLike
            The filepath or array to validate as a supported type.
        fs_kwargs: Dict[str, Any]
            Any specific keyword arguments to pass down to the fsspec created
            filesystem.
            Default: {}
        kwargs: Any
            Any kwargs used for reading and validation of the file.

        Returns
        -------
        supported: bool
            Boolean indicated if the provided data is or is not supported by the
            current Reader.

        Raises
        ------
        TypeError
            Invalid type provided to image parameter.
        """
        # Check path
        if isinstance(image, (str, Path)):
            # Expand details of provided image
            fs, path = pathlike_to_fs(
                image,
                enforce_exists=True,
                fs_kwargs=fs_kwargs,
            )

            return cls._is_supported_image(fs, path, **kwargs)

        # Special cases
        if isinstance(image, (list, np.ndarray, da.core.Array, xr.DataArray)):
            return cls._is_supported_image(image, **kwargs)

        # Raise because none of the above returned
        raise TypeError(
            f"Reader only accepts types: {types.ImageLike}. Received: '{type(image)}'."
        )

    def __init__(self, image: Any, **kwargs: Any):
        pass

    @staticmethod
    def _guess_dim_order(shape: Tuple[int, ...]) -> str:
        """
        Given an image shape attempts to guess the dimension order.

        Can and should be overwritten by inhereting Reader classes to make more
        informed guesses based off the format when possible.

        Parameters
        ----------
        shape: Tuple[int, ...]
            Tuple of the image array's dimensions.

        Returns
        -------
        dim_order: str
            The guessed dimension order.
        """
        return DEFAULT_DIMENSION_ORDER[len(DEFAULT_DIMENSION_ORDER) - len(shape) :]

    @property
    @abstractmethod
    def scenes(self) -> Tuple[str, ...]:
        """
        Returns
        -------
        scenes: Tuple[str, ...]
            A tuple of valid scene ids in the file.

        Notes
        -----
        Scene IDs are strings - not a range of integers.

        When iterating over scenes please use:

        >>> for id in image.scenes

        and not:

        >>> for i in range(len(image.scenes))
        """

    @property
    def current_scene(self) -> str:
        """
        Returns
        -------
        scene: str
            The current operating scene.
        """
        return self.scenes[self._current_scene_index]

    @property
    def current_scene_index(self) -> int:
        """
        Returns
        -------
        scene_index: int
            The current operating scene index in the file.
        """
        return self._current_scene_index

    @property
    def resolution_levels(self) -> Tuple[int, ...]:
        """
        Returns
        -------
        resolution_levels: Tuple[str, ...]
            Return the available resolution levels for the current scene.
            By default these are ordered from highest resolution to lowest
            resolution.
        """
        return (self._current_resolution_level,)

    @property
    def current_resolution_level(self) -> int:
        """
        Returns
        -------
        resolution_level: int
            The current resolution level.
        """
        return self._current_resolution_level

    def _reset_self(self) -> None:
        # Reset the data stored in the Reader object
        self._xarray_dask_data = None
        self._xarray_data = None
        self._mosaic_xarray_dask_data = None
        self._mosaic_xarray_data = None
        self._dims = None
        self._metadata = None

    def set_scene(self, scene_id: Union[str, int]) -> None:
        """
        Set the operating scene.

        Parameters
        ----------
        scene_id: Union[str, int]
            The scene id (if string) or scene index (if integer)
            to set as the operating scene.

        Raises
        ------
        IndexError
            The provided scene id or index is not found in the available scene id list.
        TypeError
            The provided value wasn't a string (scene id) or integer (scene index).
        """
        # Route to int or str setting
        if isinstance(scene_id, str):
            # Only need to run when the scene id is different from current scene
            if scene_id != self.current_scene:
                # Validate scene id
                if scene_id not in self.scenes:
                    raise IndexError(
                        f"Scene id: '{scene_id}' "
                        f"is not present in available image scenes: {self.scenes}"
                    )

                # Update current scene
                self._current_scene_index = self.scenes.index(scene_id)

                # Reset self for future read
                self._reset_self()

        # Handle index
        elif isinstance(scene_id, int):
            # Only need to run when scene index is different from current scene
            if scene_id != self.current_scene_index:
                # Validate scene index
                if scene_id >= len(self.scenes):
                    raise IndexError(
                        f"Scene index: {scene_id} "
                        f"is greater than the maximum available scene index "
                        f"({len(self.scenes) - 1}) present in the file."
                    )

                # Update current scene
                self._current_scene_index = scene_id

                # Reset set for future read
                self._reset_self()

        else:
            raise TypeError(
                f"Must provide either a string (for scene id) "
                f"or integer (for scene index). Provided: {scene_id} ({type(scene_id)}."
            )

    def set_resolution_level(self, resolution_level: int) -> None:
        """
        Set the resolution level.

        Parameters
        ----------
        resolution_level: int
            The resolution level to access the image at.

        Raises
        ------
        IndexError
            The provided resolution level is not found in the
            available resolution level list.
        """
        # Validate resolution level
        if resolution_level not in self.resolution_levels:
            raise IndexError(
                f"Resolution level: '{resolution_level}' "
                "is not present in available image resolution levels: "
                f"{self.resolution_levels}. Readers are not required by `bioio-base` "
                "to support resolution levels and therefore may not have any "
                "available besides the default of 0."
            )

        if resolution_level != self._current_resolution_level:
            self._current_resolution_level = resolution_level
            # Reset self for future read
            # Note that this resets metadata as well.
            self._reset_self()

    @abstractmethod
    def _read_delayed(self) -> xr.DataArray:
        """
        The delayed data array constructor for the image.

        Returns
        -------
        data: xr.DataArray
            The fully constructed delayed DataArray.

            It is additionally recommended to closely monitor how dask array chunks are
            managed.

        Notes
        -----
        Requirements for the returned xr.DataArray:
        * Must have the `dims` populated.
        * If a channel dimension is present, please populate the channel dimensions
        coordinate array the respective channel coordinate values.
        """

    @abstractmethod
    def _read_immediate(self) -> xr.DataArray:
        """
        The immediate data array constructor for the image.

        Returns
        -------
        data: xr.DataArray
            The fully read data array.

        Notes
        -----
        Requirements for the returned xr.DataArray:
        * Must have the `dims` populated.
        * If a channel dimension is present, please populate the channel dimensions
        coordinate array the respective channel coordinate values.
        """

    def _get_stitched_dask_mosaic(self) -> xr.DataArray:
        """
        Stitch all mosaic tiles back together and return as a single xr.DataArray with
        a delayed dask array for backing data.

        Returns
        -------
        mosaic: xr.DataArray
            The fully stitched together image. Contains all the dimensions of the image
            with the YX expanded to the full mosaic.

        Raises
        ------
        NotImplementedError
            Reader or format doesn't support reconstructing mosaic tiles.

        Notes
        -----
        Implementers can determine how to chunk the array.
        Most common is to chunk by tile.
        """
        raise NotImplementedError(
            "This reader does not support reconstructing mosaic images."
        )

    def _get_stitched_mosaic(self) -> xr.DataArray:
        """
        Stitch all mosaic tiles back together and return as a single xr.DataArray with
        an in-memory numpy array for backing data.

        Returns
        -------
        mosaic: np.ndarray
            The fully stitched together image. Contains all the dimensions of the image
            with the YX expanded to the full mosaic.

        Raises
        ------
        NotImplementedError
            Reader or format doesn't support reconstructing mosaic tiles.
        """
        raise NotImplementedError(
            "This reader does not support reconstructing mosaic images."
        )

    @property
    def xarray_dask_data(self) -> xr.DataArray:
        """
        Returns
        -------
        xarray_dask_data: xr.DataArray
            The delayed image and metadata as an annotated data array.
        """
        if self._xarray_dask_data is None:
            self._xarray_dask_data = self._read_delayed()

        return self._xarray_dask_data

    @property
    def xarray_data(self) -> xr.DataArray:
        """
        Returns
        -------
        xarray_data: xr.DataArray
            The fully read image and metadata as an annotated data array.
        """
        if self._xarray_data is None:
            self._xarray_data = self._read_immediate()

            # Remake the delayed xarray dataarray object using a rechunked dask array
            # from the just retrieved in-memory xarray dataarray
            self._xarray_dask_data = xr.DataArray(
                da.from_array(self._xarray_data.data),
                dims=self._xarray_data.dims,
                coords=self._xarray_data.coords,
                attrs=self._xarray_data.attrs,
            )

        return self._xarray_data

    @property
    def dask_data(self) -> da.Array:
        """
        Returns
        -------
        dask_data: da.Array
            The image as a dask array with the native dimension ordering.
        """
        return self.xarray_dask_data.data

    @property
    def data(self) -> np.ndarray:
        """
        Returns
        -------
        data: np.ndarray
            The image as a numpy array with native dimension ordering.
        """
        return self.xarray_data.data

    @property
    def mosaic_xarray_dask_data(self) -> xr.DataArray:
        """
        Returns
        -------
        xarray_dask_data: xr.DataArray
            The delayed mosaic image and metadata as an annotated data array.

        Raises
        ------
        InvalidDimensionOrderingError
            No MosaicTile dimension available to reader.

        Notes
        -----
        Each reader can implement mosaic tile stitching differently but it is common
        that each tile is a dask array chunk.
        """
        # Catch non-mosaic images
        if DimensionNames.MosaicTile not in self.dims.order:
            raise exceptions.InvalidDimensionOrderingError(
                "Cannot create stitched mosaic image for array without tiles available."
            )

        # Stitch, store, and return
        if self._mosaic_xarray_dask_data is None:
            self._mosaic_xarray_dask_data = self._get_stitched_dask_mosaic()

        return self._mosaic_xarray_dask_data

    @property
    def mosaic_xarray_data(self) -> xr.DataArray:
        """
        Returns
        -------
        xarray_dask_data: xr.DataArray
            The in-memory mosaic image and metadata as an annotated data array.

        Raises
        ------
        InvalidDimensionOrderingError
            No MosaicTile dimension available to reader.

        Notes
        -----
        Very large images should use `mosaic_xarray_dask_data` to avoid seg-faults.
        """
        # Catch non-mosaic images
        if DimensionNames.MosaicTile not in self.dims.order:
            raise exceptions.InvalidDimensionOrderingError(
                "Cannot create stitched mosaic image for array without tiles available."
            )

        # Stitch, store, and return
        if self._mosaic_xarray_data is None:
            self._mosaic_xarray_data = self._get_stitched_mosaic()

        return self._mosaic_xarray_data

    @property
    def mosaic_dask_data(self) -> da.Array:
        """
        Returns
        -------
        dask_data: da.Array
            The stitched together mosaic image as a dask array.

        Raises
        ------
        InvalidDimensionOrderingError
            No MosaicTile dimension available to reader.

        Notes
        -----
        Each reader can implement mosaic tile stitching differently but it is common
        that each tile is a dask array chunk.
        """
        return self.mosaic_xarray_dask_data.data

    @property
    def mosaic_data(self) -> np.ndarray:
        """
        Returns
        -------
        data: np.ndarray
            The stitched together mosaic image as a numpy array.

        Raises
        ------
        InvalidDimensionOrderingError
            No MosaicTile dimension available to reader.

        Notes
        -----
        Very large images should use `mosaic_dask_data` to avoid seg-faults.
        """
        return self.mosaic_xarray_data.data

    @property
    def dtype(self) -> np.dtype:
        """
        Returns
        -------
        dtype: np.dtype
            Data-type of the image array's elements.
        """
        return self.xarray_dask_data.dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Returns
        -------
        shape: Tuple[int, ...]
            Tuple of the image array's dimensions.
        """
        return self.xarray_dask_data.shape

    @property
    def dims(self) -> Dimensions:
        """
        Returns
        -------
        dims: Dimensions
            Object with the paired dimension names and their sizes.
        """
        if self._dims is None:
            self._dims = Dimensions(dims=self.xarray_dask_data.dims, shape=self.shape)

        return self._dims

    def get_image_dask_data(
        self, dimension_order_out: Optional[str] = None, **kwargs: Any
    ) -> da.Array:
        """
        Get specific dimension image data out of an image as a dask array.

        Parameters
        ----------
        dimension_order_out: Optional[str]
            A string containing the dimension ordering desired for the returned ndarray.
            Default: The natural image dimension order.

        kwargs: Any
            * C=1: specifies Channel 1
            * T=3: specifies the fourth index in T
            * D=n: D is Dimension letter and n is the index desired. D should not be
              present in the dimension_order_out.
            * D=[a, b, c]: D is Dimension letter and a, b, c is the list of indices
              desired. D should be present in the dimension_order_out.
            * D=(a, b, c): D is Dimension letter and a, b, c is the tuple of indices
              desired. D should be present in the dimension_order_out.
            * D=range(...): D is Dimension letter and range is the standard Python
              range function. D should be present in the dimension_order_out.
            * D=slice(...): D is Dimension letter and slice is the standard Python
              slice function. D should be present in the dimension_order_out.

        Returns
        -------
        data: da.Array
            The image data with the specified dimension ordering.

        Examples
        --------
        Specific index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... c1 = img.get_image_dask_data("ZYX", C=1)

        List of index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_second = img.get_image_dask_data("CZYX", C=[0, 1])

        Tuple of index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_last = img.get_image_dask_data("CZYX", C=(0, -1))

        Range of index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_three = img.get_image_dask_data("CZYX", C=range(3))

        Slice selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... every_other = img.get_image_dask_data("CZYX", C=slice(0, -1, 2))

        Notes
        -----
        If a requested dimension is not present in the data the dimension is
        added with a depth of 1.

        See `aicsimageio.transforms.reshape_data` for more details.
        """
        # If no out orientation, simply return current data as dask array
        if dimension_order_out is None:
            return self.dask_data

        # Transform and return
        return transforms.reshape_data(
            data=self.dask_data,
            given_dims=self.dims.order,
            return_dims=dimension_order_out,
            **kwargs,
        )

    def get_image_data(
        self, dimension_order_out: Optional[str] = None, **kwargs: Any
    ) -> np.ndarray:
        """
        Read the image as a numpy array then return specific dimension image data.

        Parameters
        ----------
        dimension_order_out: Optional[str]
            A string containing the dimension ordering desired for the returned ndarray.
            Default: The natural image dimension order.

        kwargs: Any
            * C=1: specifies Channel 1
            * T=3: specifies the fourth index in T
            * D=n: D is Dimension letter and n is the index desired. D should not be
              present in the dimension_order_out.
            * D=[a, b, c]: D is Dimension letter and a, b, c is the list of indices
              desired. D should be present in the dimension_order_out.
            * D=(a, b, c): D is Dimension letter and a, b, c is the tuple of indices
              desired. D should be present in the dimension_order_out.
            * D=range(...): D is Dimension letter and range is the standard Python
              range function. D should be present in the dimension_order_out.
            * D=slice(...): D is Dimension letter and slice is the standard Python
              slice function. D should be present in the dimension_order_out.

        Returns
        -------
        data: np.ndarray
            The image data with the specified dimension ordering.

        Examples
        --------
        Specific index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... c1 = img.get_image_data("ZYX", C=1)

        List of index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_second = img.get_image_data("CZYX", C=[0, 1])

        Tuple of index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_and_last = img.get_image_data("CZYX", C=(0, -1))

        Range of index selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... first_three = img.get_image_dask_data("CZYX", C=range(3))

        Slice selection

        >>> img = Reader("s_1_t_1_c_10_z_20.ome.tiff")
        ... every_other = img.get_image_data("CZYX", C=slice(0, -1, 2))

        Notes
        -----
        * If a requested dimension is not present in the data the dimension is
          added with a depth of 1.
        * This will preload the entire image before returning the requested data.

        See `aicsimageio.transforms.reshape_data` for more details.
        """
        # If no out orientation, simply return current data as dask array
        if dimension_order_out is None:
            return self.data

        # Transform and return
        return transforms.reshape_data(
            data=self.data,
            given_dims=self.dims.order,
            return_dims=dimension_order_out,
            **kwargs,
        )

    @property
    def metadata(self) -> Any:
        """
        Returns
        -------
        metadata: Any
            The metadata for the formats supported by the inhereting Reader.

            If the inheriting Reader supports processing the metadata into a more useful
            format / Python object, this will return the result.

            For both the unprocessed and processed metadata from the file, use
            `xarray_dask_data.attrs` which will contain a dictionary with keys:
            `unprocessed` and `processed` that you can then select.
        """
        if self._metadata is None:
            if constants.METADATA_PROCESSED in self.xarray_dask_data.attrs:
                self._metadata = self.xarray_dask_data.attrs[
                    constants.METADATA_PROCESSED
                ]
            else:
                self._metadata = self.xarray_dask_data.attrs[
                    constants.METADATA_UNPROCESSED
                ]

        return self._metadata

    @property
    def ome_metadata(self) -> OME:
        """
        Returns
        -------
        metadata: OME
            The original metadata transformed into the OME specfication.
            This likely isn't a complete transformation but is guarenteed to
            be a valid transformation.

        Raises
        ------
        NotImplementedError
            No metadata transformer available.
        """
        raise NotImplementedError()

    @property
    def channel_names(self) -> Optional[List[str]]:
        """
        Returns
        -------
        channel_names: List[str]
            Using available metadata, the list of strings representing channel names.
            If no channel dimension present in the data, returns None.
        """
        if DimensionNames.Channel in self.xarray_dask_data.dims:
            return list(self.xarray_dask_data[DimensionNames.Channel].values)

        return None

    @staticmethod
    def _generate_coord_array(
        start: Union[int, float], stop: Union[int, float], step_size: Union[int, float]
    ) -> np.ndarray:
        """
        Generate an np.ndarray for coordinate values.

        Parameters
        ----------
        start: Union[int, float]
            The start value.
        stop: Union[int, float]
            The stop value.
        step_size: Union[int, float]
            How large each step should be.

        Returns
        -------
        coords: np.ndarray
            The coordinate array.

        Notes
        -----
        In general, we have learned that floating point math is hard....
        This block of code used to use `np.arange` with floats as parameters and
        it was causing errors. To solve, we generate the range with ints and then
        multiply by a float across the entire range to get the proper coords.
        See: https://github.com/AllenCellModeling/aicsimageio/issues/249
        """
        return np.arange(start, stop) * step_size

    @property
    def physical_pixel_sizes(self) -> PhysicalPixelSizes:
        """
        Returns
        -------
        sizes: PhysicalPixelSizes
            Using available metadata, the floats representing physical pixel sizes for
            dimensions Z, Y, and X.

        Notes
        -----
        We currently do not handle unit attachment to these values. Please see the file
        metadata for unit information.
        """
        return PhysicalPixelSizes(None, None, None)

    def get_mosaic_tile_position(
        self, mosaic_tile_index: int, **kwargs: int
    ) -> Tuple[int, int]:
        """
        Get the absolute position of the top left point for a single mosaic tile.

        Parameters
        ----------
        mosaic_tile_index: int
            The index for the mosaic tile to retrieve position information for.
        kwargs: int
            The keywords below allow you to specify the dimensions that you wish
            to match. If you under-specify the constraints you can easily
            end up with a massive image stack.
                       Z = 1   # The Z-dimension.
                       C = 2   # The C-dimension ("channel").
                       T = 3   # The T-dimension ("time").

        Returns
        -------
        top: int
            The Y coordinate for the tile position.
        left: int
            The X coordinate for the tile position.

        Raises
        ------
        UnexpectedShapeError
            The image has no mosaic dimension available.
        IndexError
            No matching mosaic tile index found.
        """
        raise NotImplementedError()

    def get_mosaic_tile_positions(self, **kwargs: int) -> List[Tuple[int, int]]:
        """
        Get the absolute positions of the top left points for each mosaic tile
        matching the specified dimensions and current scene.

        Parameters
        ----------
        kwargs: int
            The keywords below allow you to specify the dimensions that you wish
            to match. If you under-specify the constraints you can easily
            end up with a massive image stack.
                       Z = 1   # The Z-dimension.
                       C = 2   # The C-dimension ("channel").
                       T = 3   # The T-dimension ("time").

        Returns
        -------
        mosaic_tile_positions: List[Tuple[int, int]]
            List of the Y and X coordinate for the tile positions.

        Raises
        ------
        UnexpectedShapeError
            The image has no mosaic dimension available.
        """
        raise NotImplementedError()

    @property
    def mosaic_tile_dims(self) -> Optional[Dimensions]:
        """
        Returns
        -------
        tile_dims: Optional[Dimensions]
            The dimensions for each tile in the mosaic image.
            If the image is not a mosaic image, returns None.
        """
        if DimensionNames.MosaicTile in self.dims.order:
            return Dimensions("YX", (self.dims.Y, self.dims.X))

        return None

    def get_stack(self, **kwargs: Any) -> np.ndarray:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: np.ndarray
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        aicsimageio.transforms.generate_stack:
            Underlying function for generating various scene stacks.
        """
        return transforms.generate_stack(self, mode="data", **kwargs)

    def get_dask_stack(self, **kwargs: Any) -> da.Array:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: da.Array
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        aicsimageio.transforms.generate_stack:
            Underlying function for generating various scene stacks.
        """
        return transforms.generate_stack(self, mode="dask_data", **kwargs)

    def get_xarray_stack(self, **kwargs: Any) -> xr.DataArray:
        """
        Get all scenes stacked in to a single array.


        Returns
        -------
        stack: xr.DataArray
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        aicsimageio.transforms.generate_stack:
            Underlying function for generating various scene stacks.

        Notes
        -----
        When requesting an xarray stack, the first scene's coordinate planes
        are used for the returned xarray DataArray object coordinate planes.
        """
        return transforms.generate_stack(self, mode="xarray_data", **kwargs)

    def get_xarray_dask_stack(self, **kwargs: Any) -> xr.DataArray:
        """
        Get all scenes stacked in to a single array.

        Returns
        -------
        stack: xr.DataArray
            The fully stacked array. This can be 6+ dimensions with Scene being
            the first dimension.
        kwargs: Any
            Extra keyword arguments that will be passed down to the
            generate stack function.

        See Also
        --------
        aicsimageio.transforms.generate_stack:
            Underlying function for generating various scene stacks.

        Notes
        -----
        When requesting an xarray stack, the first scene's coordinate planes
        are used for the returned xarray DataArray object coordinate planes.
        """
        return transforms.generate_stack(self, mode="xarray_dask_data", **kwargs)

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"[Image-is-in-Memory: {self._xarray_data is not None}]>"
        )

    def __repr__(self) -> str:
        return str(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Tuple

from fsspec.core import url_to_fs

if TYPE_CHECKING:
    from fsspec.spec import AbstractFileSystem

from .types import PathLike

###############################################################################


def pathlike_to_fs(
    uri: PathLike,
    enforce_exists: bool = False,
    fs_kwargs: Dict[str, Any] = {},
) -> Tuple["AbstractFileSystem", str]:
    """
    Find and return the appropriate filesystem and path from a path-like object.

    Parameters
    ----------
    uri: PathLike
        The local or remote path or uri.
    enforce_exists: bool
        Check whether or not the resource exists, if not, raise FileNotFoundError.

    Returns
    -------
    fs: AbstractFileSystem
        The filesystem to operate on.
    path: str
        The full path to the target resource.
    fs_kwargs: Dict[str, Any]
        Any specific keyword arguments to pass down to the fsspec created filesystem.
        Default: {}

    Raises
    ------
    FileNotFoundError
        If enforce_exists is provided value True and the resource is not found or is
        unavailable.
    """
    # Convert paths to string to be handled by url_to_fs
    if isinstance(uri, Path):
        uri = str(uri)

    # Get details
    fs, path = url_to_fs(uri, **fs_kwargs)

    # Check file exists
    if enforce_exists:
        if not fs.exists(path):
            raise FileNotFoundError(f"{fs.protocol}://{path}")

    # Get and store details
    # We do not return an AbstractBufferedFile (i.e. fs.open) as we do not want to have
    # any open file buffers _after_ any API call. API calls must themselves call
    # fs.open and complete their function during the context of the opened buffer.
    return fs, path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections.abc import Sequence as seq
from typing import Collection, ItemsView, Sequence, Tuple, Union

###############################################################################


class DimensionNames:
    Time = "T"
    Channel = "C"
    SpatialZ = "Z"
    SpatialY = "Y"
    SpatialX = "X"
    Samples = "S"
    MosaicTile = "M"


DEFAULT_DIMENSION_ORDER_LIST = [
    DimensionNames.Time,
    DimensionNames.Channel,
    DimensionNames.SpatialZ,
    DimensionNames.SpatialY,
    DimensionNames.SpatialX,
]
DEFAULT_DIMENSION_ORDER_LIST_WITH_SAMPLES = DEFAULT_DIMENSION_ORDER_LIST + [
    DimensionNames.Samples
]
DEFAULT_DIMENSION_ORDER_LIST_WITH_MOSAIC_TILES = [
    DimensionNames.MosaicTile
] + DEFAULT_DIMENSION_ORDER_LIST
DEFAULT_DIMENSION_ORDER_LIST_WITH_MOSAIC_TILES_AND_SAMPLES = (
    [DimensionNames.MosaicTile]
    + DEFAULT_DIMENSION_ORDER_LIST
    + [DimensionNames.Samples]
)

DEFAULT_DIMENSION_ORDER = "".join(DEFAULT_DIMENSION_ORDER_LIST)
DEFAULT_DIMENSION_ORDER_WITH_SAMPLES = "".join(
    DEFAULT_DIMENSION_ORDER_LIST_WITH_SAMPLES
)
DEFAULT_DIMENSION_ORDER_WITH_MOSAIC_TILES = "".join(
    DEFAULT_DIMENSION_ORDER_LIST_WITH_MOSAIC_TILES
)
DEFAULT_DIMENSION_ORDER_WITH_MOSAIC_TILES_AND_SAMPLES = "".join(
    DEFAULT_DIMENSION_ORDER_LIST_WITH_MOSAIC_TILES_AND_SAMPLES
)

DEFAULT_CHUNK_DIMS = [
    DimensionNames.SpatialZ,
    DimensionNames.SpatialY,
    DimensionNames.SpatialX,
    DimensionNames.Samples,
]

REQUIRED_CHUNK_DIMS = [
    DimensionNames.SpatialY,
    DimensionNames.SpatialX,
    DimensionNames.Samples,
]

###############################################################################


class Dimensions:
    def __init__(self, dims: Collection[str], shape: Tuple[int, ...]):
        """
        A general object for managing the pairing of dimension name and dimension size.

        Parameters
        ----------
        dims: Collection[str]
            An ordered string or collection of the dimensions to pair with their sizes.
        shape: Tuple[int, ...]
            An ordered tuple of the dimensions sizes to pair with their names.

        Examples
        --------
        >>> dims = Dimensions("TCZYX", (1, 4, 75, 624, 924))
        ... dims.X
        ... dims['T', 'X']
        """
        # zip(strict=True) only in Python 3.10
        # Check equal length dims and shape
        if len(dims) != len(shape):
            raise ValueError(
                f"Number of dimensions provided ({len(dims)} -- '{dims}') "
                f"does not match shape size provided ({len(shape)} -- '{shape}')."
            )

        # Make dims a string
        if not isinstance(dims, str):
            if any([len(c) != 1 for c in dims]):
                raise ValueError(
                    f"When providing a list of dimension strings, "
                    f"each dimension may only be a single character long "
                    f"(received: '{dims}')."
                )
            dims = "".join(dims)

        # Store order and shape
        self._order = dims
        self._shape = shape

        # Create attributes
        self._dims_shape = dict(zip(dims, shape))
        for dim, size in self._dims_shape.items():
            setattr(self, dim, size)

    @property
    def order(self) -> str:
        """
        Returns
        -------
        order: str
            The natural order of the dimensions as a single string.
        """
        return self._order

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Returns
        -------
        shape: Tuple[int, ...]
            The dimension sizes in their natural order.
        """
        return self._shape

    def items(self) -> ItemsView[str, int]:
        return self._dims_shape.items()

    def __str__(self) -> str:
        dims_string = ", ".join([f"{dim}: {size}" for dim, size in self.items()])
        return f"<Dimensions [{dims_string}]>"

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, key: Union[str, Sequence[str]]) -> Tuple[int, ...]:
        if isinstance(key, str):
            if key not in self._order:
                raise IndexError(f"{key} not in {self._order}")
            return (self._dims_shape[key],)
        elif isinstance(key, seq) and all(isinstance(k, str) for k in key):
            invalid_dims = []
            for k in key:
                if k not in self._order:
                    invalid_dims.append(k)
            if len(invalid_dims) == 0:
                return tuple(self._dims_shape[k] for k in key)
            else:
                raise IndexError(f"{', '.join(invalid_dims)} not in {self._order}")
        else:
            raise TypeError(
                f"Key must be a string or list of strings but got type {type(key)}"
            )

    def __setattr__(self, __name: str, __value: int) -> None:
        super().__setattr__(__name, __value)

    def __getattr__(self, __name: str) -> int:
        # TODO: Py310 __match_args__ for better typing
        return super().__getattribute__(__name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from collections import Counter
from numbers import Integral
from typing import Any, List, Literal, Optional, Tuple, Union

import dask.array as da
import numpy as np
import xarray as xr

from . import types
from .exceptions import ConflictingArgumentsError, UnexpectedShapeError
from .image_container import ImageContainer

###############################################################################


def reduce_to_slice(L: Union[List, Tuple]) -> Union[int, List, slice, Tuple]:
    # if the list only has one element, then just use it
    if len(L) == 1:
        return slice(L[0], L[0] + 1)
    # if the list has at least 2 elements we can check for sliceable
    # it is convertable to a slice if the step size between each
    # consecutive pair of elements is equal and positive
    # 1. get all the deltas in a list:
    steps = [(L[i + 1] - L[i]) for i in range(len(L) - 1)]
    # 2. check if all the deltas are equal and positive
    if steps[0] > 0 and steps.count(steps[0]) == len(steps):
        return slice(min(L), max(L) + 1, steps[0])
    # if we can't convert to a slice, then just return the list unmodified
    return L


def transpose_to_dims(
    data: types.ArrayLike,
    given_dims: str,
    return_dims: str,
) -> types.ArrayLike:
    """
    This shuffles the data dimensions from given_dims to return_dims. Each dimension
    must be present in given_dims must be used in return_dims

    Parameters
    ----------
    data: types.ArrayLike
        Either a dask array or numpy.ndarray of arbitrary shape but with the dimensions
        specified in given_dims
    given_dims: str
        The dimension ordering of data, "CZYX", "VBTCXZY" etc
    return_dims: str
        The dimension ordering of the return data

    Returns
    -------
    data: types.ArrayLike
        The data with the specified dimension ordering.

    Raises
    ------
    ConflictingArgumentsError
        given_dims and return_dims are incompatible.
    """
    # Use a counter to track that the contents are composed of the same letters
    # and that no letter is repeated
    if (
        Counter(given_dims) != Counter(return_dims)
        or max(Counter(given_dims).values()) > 1
    ):
        raise ConflictingArgumentsError(
            f"given_dims={given_dims} and return_dims={return_dims} are incompatible."
        )

    # Resort the data into return_dims order
    match_map = {dim: given_dims.find(dim) for dim in given_dims}
    transposer = []
    for dim in return_dims:
        transposer.append(match_map[dim])
    data = data.transpose(transposer)

    return data


def reshape_data(
    data: types.ArrayLike, given_dims: str, return_dims: str, **kwargs: Any
) -> types.ArrayLike:
    """
    Reshape the data into return_dims, pad missing dimensions, and prune extra
    dimensions. Warns the user to use the base reader if the depth of the Dimension
    being removed is not 1.

    Parameters
    ----------
    data: types.ArrayLike
        Either a dask array or numpy.ndarray of arbitrary shape but with the dimensions
        specified in given_dims
    given_dims: str
        The dimension ordering of data, "CZYX", "VBTCXZY" etc
    return_dims: str
        The dimension ordering of the return data
    kwargs:
        * C=1 => desired specific channel, if C in the input data has depth 3 then C=1
          returns the 2nd slice (0 indexed)
        * Z=10 => desired specific channel, if Z in the input data has depth 20 then
          Z=10 returns the 11th slice
        * T=[0, 1] => desired specific timepoints, if T in the input data has depth 100
          then T=[0, 1] returns the 1st and 2nd slice (0 indexed)
        * T=(0, 1) => desired specific timepoints, if T in the input data has depth 100
          then T=(0, 1) returns the 1st and 2nd slice (0 indexed)
        * T=(0, -1) => desired specific timepoints, if T in the input data has depth 100
          then T=(0, -1) returns the first and last slice
        * T=range(10) => desired specific timepoints, if T in the input data has depth
          100 then T=range(10) returns the first ten slices
        * T=slice(0, -1, 5) => desired specific timepoints, T=slice(0, -1, 5) returns
          every fifth timepoint

    Returns
    -------
    data: types.ArrayLike
        The data with the specified dimension ordering.

    Raises
    ------
    ConflictingArgumentsError
        Missing dimension in return dims when using range, slice, or multi-index
        dimension selection for the requested dimension.

    IndexError
        Requested dimension index not present in data.

    Examples
    --------
    Specific index selection

    >>> data = np.random.rand((10, 100, 100))
    ... z1 = reshape_data(data, "ZYX", "YX", Z=1)

    List of index selection

    >>> data = np.random.rand((10, 100, 100))
    ... first_and_second = reshape_data(data, "ZYX", "YX", Z=[0, 1])

    Tuple of index selection

    >>> data = np.random.rand((10, 100, 100))
    ... first_and_last = reshape_data(data, "ZYX", "YX", Z=(0, -1))

    Range of index selection

    >>> data = np.random.rand((10, 100, 100))
    ... first_three = reshape_data(data, "ZYX", "YX", Z=range(3))

    Slice selection

    >>> data = np.random.rand((10, 100, 100))
    ... every_other = reshape_data(data, "ZYX", "YX", Z=slice(0, -1, 2))

    Empty dimension expansion

    >>> data = np.random.rand((10, 100, 100))
    ... with_time = reshape_data(data, "ZYX", "TZYX")

    Dimension order shuffle

    >>> data = np.random.rand((10, 100, 100))
    ... as_zx_base = reshape_data(data, "ZYX", "YZX")

    Selections, empty dimension expansions, and dimension order shuffle

    >>> data = np.random.rand((10, 100, 100))
    ... example = reshape_data(data, "CYX", "BSTCZYX", C=slice(0, -1, 3))
    """
    # Check for parameter conflicts
    for dim in given_dims:
        # return_dims='CZYX' and iterable dimensions 'T=range(10)'
        # Dimension is in kwargs
        # Dimension is an iterable
        # Dimension is not in return dimensions
        if (
            isinstance(kwargs.get(dim), (list, tuple, range, slice))
            and dim not in return_dims
        ):
            raise ConflictingArgumentsError(
                f"When selecting a multiple dimension indices, the specified "
                f"dimension must be provided in return_dims. "
                f"return_dims={return_dims}, dimension {dim} = {kwargs.get(dim)}"
            )

    # Process each dimension available
    new_dims = given_dims
    dim_specs = []
    for dim in given_dims:
        # Store index of the dim as it is in given data
        dim_index = given_dims.index(dim)

        # Handle dim in return_dims which means that it is
        # an iterable or None selection
        if dim in return_dims:
            # Specific iterable requested
            if dim in kwargs:
                # Actual dim specification
                # The specification provided for this dimension in the kwargs
                dim_spec = kwargs.get(dim)
                display_dim_spec = dim_spec

                if isinstance(dim_spec, int):
                    dim_spec = slice(dim_spec, dim_spec + 1)

                # Convert operator to standard list or slice
                # dask.Array and numpy.ndarray both natively support
                # List[int] and slices being passed to getitem so no need to cast them
                # to anything different
                if isinstance(dim_spec, (tuple, range)):
                    dim_spec = list(dim_spec)

                # Get the largest absolute value index in the list using min and max
                if isinstance(dim_spec, list):
                    check_selection_max = max([abs(min(dim_spec)), max(dim_spec)])
                    # try to convert to slice if possible
                    dim_spec = reduce_to_slice(dim_spec)

                # Get the largest absolute value index from start and stop of slice
                if isinstance(dim_spec, slice):
                    check_selection_max = max([abs(dim_spec.stop), abs(dim_spec.start)])
            else:
                # Nothing was requested from this dimension
                dim_spec = slice(None, None, None)
                display_dim_spec = dim_spec

                # No op means that it doesn't matter how much data is in this dimension
                check_selection_max = 0

        # Not in return_dims means that it is a fixed integer selection
        else:
            if dim in kwargs:
                # Integer requested
                dim_spec = kwargs.get(dim)
                display_dim_spec = dim_spec

                # Check that integer
                if not isinstance(dim_spec, Integral):
                    raise TypeError(
                        "Dimensions not in output must be integers. "
                        f"Got {type(dim_spec).__name__} for {dim}."
                    )
                check_selection_max = dim_spec
            else:
                dim_spec = 0
                display_dim_spec = dim_spec
                check_selection_max = 0

            # Remove dim from new dims as it is fixed size
            new_dims = new_dims.replace(dim, "")

        # Check that fixed integer request isn't outside of request
        if check_selection_max > data.shape[dim_index]:
            raise IndexError(
                f"Dimension specified with {dim}={display_dim_spec} "
                f"but Dimension shape is {data.shape[dim_index]}."
            )

        # All checks and operations passed, append dim operation to getitem ops
        dim_specs.append(dim_spec)

    # Run getitems
    data = data[tuple(dim_specs)]

    # Add empty dims where dimensions were requested but data doesn't exist
    # Add dimensions to new dims where empty dims are added
    for i, dim in enumerate(return_dims):
        # This dimension wasn't processed
        if dim not in given_dims:
            new_dims = f"{new_dims[:i]}{dim}{new_dims[i:]}"
            data = data.reshape(*data.shape[:i], 1, *data.shape[i:])

    # Any extra dimensions have been removed, only a problem if the depth is > 1
    return transpose_to_dims(
        data, given_dims=new_dims, return_dims=return_dims
    )  # don't pass kwargs or 2 copies


def generate_stack(
    image_container: ImageContainer,
    mode: Literal["data", "dask_data", "xarray_data", "xarray_dask_data"],
    drop_non_matching_scenes: bool = False,
    select_scenes: Optional[
        Union[list[Union[str, int]], tuple[Union[str, int], ...]]
    ] = None,
    scene_character: str = "I",
    scene_coord_values: str = "index",
) -> types.MetaArrayLike:
    """
    Stack each scene contained in the reader into a
    single array. This method handles the logic of determining which
    stack function to use (dask or numpy) and whether or not to return a
    labelled array (xr.DataArray). Users should prefer
    to use one of get_stack, get_dask_stack, get_xarray_stack, or
    get_xarray_dask_stack.

    Parameters
    ----------
    mode: Literal["data", "dask_data", "xarray_data", "xarray_dask_data"]
        String describing the style of data to return. Should be one of:
        "data", "dask_data", "xarray_data", "xarray_dask_data".
    drop_non_matching_scenes: bool
        During the scene iteration process, if the next scene to be added
        to the stack has different shape or dtype, should it be dropped or
        raise an error.  Default: False (raise an error)
    select_scenes: Optional[
            Union[List[Union[str, int]], Tuple[Union[str, int], ...]]]
        Which scenes to stack into a single array. Scenes can be provided
        as a list or tuple of scene indices or names. It is recommended to
        use the scene integer index instead of the scene name to avoid
        duplicate scene name lookup issues.
        Default: None (stack all scenes)
    scene_character: str
        Character to use as the name of the scene dimension on the output
        array. Default "I"
    scene_coord_values : str
        How to assign coordinates to the scene dimension of the final
        array. If scene_coord_values="names" use the scene name from
        the reader object. If scene_coord_values="index" don't attach any
        coordinates and fall back to integer values.
        Default: "index"

    Returns
    -------
    stack: types.MetaArrayLike
        The fully stacked array. This can be 6+ dimensions with Scene being
        the first dimension.

    """

    mode_check = ["data", "dask_data", "xarray_data", "xarray_dask_data"]
    if mode not in mode_check:
        raise ValueError(
            f"Invalid mode kwarg. Found {mode} but should be one of:"
            f"{', '.join(mode_check)}."
        )

    scene_stacks = []
    scene_names = []

    if select_scenes is None:
        select_scenes = list(range(len(image_container.scenes)))

    for i, s in enumerate(select_scenes):
        image_container.set_scene(s)
        data = getattr(image_container, mode)

        # Store the shape and dtype of the first scenes data
        # to check against later scenes. If returning a DataArray
        # store the coords and dims to use in the final output
        if i == 0:
            shape = data.shape
            dtype = data.dtype

            if "xarray" in mode:
                coords = dict(data.coords)
                dims = data.dims

                if scene_character in dims:
                    raise ValueError(
                        f"Provided scene dimension character '{scene_character}' "
                        f"was found in the existing dimensions of the data {dims}"
                    )

        # Check other scenes against the first scene
        else:
            if data.shape != shape:
                if not drop_non_matching_scenes:
                    raise UnexpectedShapeError(
                        f"All scenes must have same shape. Found shape"
                        f"{data.shape} in scene {s} but expected"
                        f"{shape} based on scene {select_scenes[0]}"
                    )
                else:
                    continue
            if data.dtype != dtype:
                if not drop_non_matching_scenes:
                    raise TypeError(
                        f"All scenes must have the same dtype. Found data"
                        f"with dtype {data.dtype} in scene {s} but expected"
                        f"dtype {dtype} based on scene {select_scenes[0]}"
                    )
                else:
                    continue

        scene_stacks.append(data)
        scene_names.append(image_container.current_scene)

    stack = da.stack if "dask" in mode else np.stack

    if "xarray" in mode:
        all_data = stack([x.data for x in scene_stacks])
        if scene_coord_values == "names":
            coords = {scene_character: scene_names, **coords}

        return xr.DataArray(
            all_data,
            dims=(scene_character, *dims),
            coords=coords,
        )

    else:
        return stack(scene_stacks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import dask.array as da
import numpy as np
import xarray as xr

from .dimensions import Dimensions
from .types import ImageLike, PhysicalPixelSizes

###############################################################################


class ImageContainer(ABC):
    def __init__(
        self,
        image: ImageLike,
        reader: Optional[Type["ImageContainer"]] = None,
        reconstruct_mosaic: bool = True,
        fs_kwargs: Dict[str, Any] = {},
        **kwargs: Any,
    ):
        pass

    @property
    @abstractmethod
    def scenes(self) -> Tuple[str, ...]:
        pass

    @property
    @abstractmethod
    def current_scene(self) -> str:
        pass

    @property
    @abstractmethod
    def current_scene_index(self) -> int:
        pass

    @abstractmethod
    def set_scene(self, scene_id: Union[str, int]) -> None:
        pass

    @property
    @abstractmethod
    def resolution_levels(self) -> Tuple[int, ...]:
        pass

    @property
    @abstractmethod
    def current_resolution_level(self) -> int:
        pass

    @abstractmethod
    def set_resolution_level(self, resolution_level: int) -> None:
        pass

    @property
    @abstractmethod
    def xarray_dask_data(self) -> xr.DataArray:
        pass

    @property
    @abstractmethod
    def xarray_data(self) -> xr.DataArray:
        pass

    @property
    @abstractmethod
    def dask_data(self) -> da.Array:
        pass

    @property
    @abstractmethod
    def data(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        pass

    @property
    @abstractmethod
    def shape(self) -> Tuple[int, ...]:
        pass

    @property
    @abstractmethod
    def dims(self) -> Dimensions:
        pass

    @abstractmethod
    def get_image_dask_data(
        self, dimension_order_out: Optional[str] = None, **kwargs: Any
    ) -> da.Array:
        pass

    @abstractmethod
    def get_image_data(
        self, dimension_order_out: Optional[str] = None, **kwargs: Any
    ) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def metadata(self) -> Any:
        pass

    @property
    @abstractmethod
    def channel_names(self) -> Optional[List[str]]:
        pass

    @property
    @abstractmethod
    def physical_pixel_sizes(self) -> PhysicalPixelSizes:
        pass



# -*- coding: utf-8 -*-

"""Top-level package for bioio_base."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("bioio-base")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Eva Maxfield Brown, Dan Toloudis, BioIO Contributors"
__email__ = "evamaxfieldbrown@gmail.com, danielt@alleninstitute.org"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, List, Optional, Tuple, Type, Union

import numpy as np
from distributed.protocol import deserialize, serialize
from fsspec.implementations.local import LocalFileSystem
from psutil import Process
from xarray.testing import assert_equal

from .image_container import ImageContainer
from .reader import Reader
from .types import PathLike

###############################################################################


def check_local_file_not_open(image_container: ImageContainer) -> None:
    if not hasattr(image_container, "_fs") or not hasattr(image_container, "_path"):
        if not hasattr(image_container, "reader"):
            raise ValueError(
                "Expected 'image_container' to have '_fs' and "
                "'_path' attributes or a 'reader' attribute"
            )

        image_container = image_container.reader
        if not hasattr(image_container, "_fs") or not hasattr(image_container, "_path"):
            raise ValueError(
                "Expected 'image_container' to have '_fs' and "
                "'_path' attributes or a 'reader' attribute with "
                "'_fs' and '_path' attributes"
            )

    # Check that there are no open file pointers
    if isinstance(image_container._fs, LocalFileSystem):
        proc = Process()
        assert str(image_container._path) not in [f.path for f in proc.open_files()]


def check_can_serialize_image_container(image_container: ImageContainer) -> None:
    # Dump and reconstruct
    reconstructed = deserialize(*serialize(image_container))

    # Assert primary attrs are equal
    if image_container.xarray_data is None:
        assert reconstructed.xarray_data is None
    else:
        assert_equal(image_container.xarray_data, reconstructed.xarray_data)

    if image_container.xarray_dask_data is None:
        assert reconstructed.xarray_dask_data is None
    else:
        assert_equal(image_container.xarray_dask_data, reconstructed.xarray_dask_data)


def run_image_container_checks(
    image_container: ImageContainer,
    set_scene: str,
    expected_scenes: Tuple[str, ...],
    expected_current_scene: str,
    expected_shape: Tuple[int, ...],
    expected_dtype: np.dtype,
    expected_dims_order: str,
    expected_channel_names: Optional[List[str]],
    expected_physical_pixel_sizes: Tuple[
        Optional[float], Optional[float], Optional[float]
    ],
    expected_metadata_type: Union[type, Tuple[Union[type, Tuple[Any, ...]], ...]],
    set_resolution_level: int = 0,
    expected_current_resolution_level: int = 0,
    expected_resolution_levels: Tuple[int, ...] = (0,),
) -> ImageContainer:
    """
    A general suite of tests to run against readers.
    """

    # Check serdes
    check_can_serialize_image_container(image_container)

    # Set scene
    image_container.set_scene(set_scene)

    # Check scene info
    assert image_container.scenes == expected_scenes
    assert image_container.current_scene == expected_current_scene

    # Set resolution level
    image_container.set_resolution_level(set_resolution_level)

    # Check resolution level info
    assert image_container.resolution_levels == expected_resolution_levels
    assert image_container.current_resolution_level == expected_current_resolution_level

    # Check basics
    assert image_container.shape == expected_shape
    assert image_container.dtype == expected_dtype
    assert image_container.dims.order == expected_dims_order
    assert image_container.dims.shape == expected_shape
    assert image_container.channel_names == expected_channel_names
    assert image_container.physical_pixel_sizes == expected_physical_pixel_sizes
    assert isinstance(image_container.metadata, expected_metadata_type)

    # Read different chunks
    zyx_chunk_from_delayed = image_container.get_image_dask_data("ZYX").compute()
    cyx_chunk_from_delayed = image_container.get_image_dask_data("CYX").compute()

    # Read in mem then pull chunks
    zyx_chunk_from_mem = image_container.get_image_data("ZYX")
    cyz_chunk_from_mem = image_container.get_image_data("CYX")

    # Compare chunk reads
    np.testing.assert_array_equal(
        zyx_chunk_from_delayed,
        zyx_chunk_from_mem,
    )
    np.testing.assert_array_equal(
        cyx_chunk_from_delayed,
        cyz_chunk_from_mem,
    )

    # Check that the shape and dtype are expected after reading in full
    assert image_container.data.shape == expected_shape
    assert image_container.data.dtype == expected_dtype

    # Check serdes
    check_can_serialize_image_container(image_container)

    return image_container


def run_reader_mosaic_checks(
    tiles_reader: Reader,
    stitched_reader: Reader,
    tiles_set_scene: str,
    stitched_set_scene: str,
) -> None:
    """
    A general suite of tests to run against readers that can stitch mosaic tiles.

    This tests uses in-memory numpy to compare. Test mosaics should be small enough to
    fit into memory.
    """
    # Set scenes
    tiles_reader.set_scene(tiles_set_scene)
    stitched_reader.set_scene(stitched_set_scene)

    # Get data subset
    from_tiles_stitched_data = tiles_reader.mosaic_data
    already_stitched_data = stitched_reader.data

    # Compare
    assert from_tiles_stitched_data.shape == already_stitched_data.shape
    np.testing.assert_array_equal(from_tiles_stitched_data, already_stitched_data)


def run_image_file_checks(
    ImageContainer: Type[ImageContainer],
    image: PathLike,
    set_scene: str,
    expected_scenes: Tuple[str, ...],
    expected_current_scene: str,
    expected_shape: Tuple[int, ...],
    expected_dtype: np.dtype,
    expected_dims_order: str,
    expected_channel_names: Optional[List[str]],
    expected_physical_pixel_sizes: Tuple[
        Optional[float], Optional[float], Optional[float]
    ],
    expected_metadata_type: Union[type, Tuple[Union[type, Tuple[Any, ...]], ...]],
    set_resolution_level: int = 0,
    expected_current_resolution_level: int = 0,
    expected_resolution_levels: Tuple[int, ...] = (0,),
) -> ImageContainer:
    # Init container
    image_container = ImageContainer(image, fs_kwargs=dict(anon=True))

    # Check for file pointers
    check_local_file_not_open(image_container)

    # Run array and metadata check operations
    run_image_container_checks(
        image_container=image_container,
        set_scene=set_scene,
        set_resolution_level=set_resolution_level,
        expected_scenes=expected_scenes,
        expected_current_scene=expected_current_scene,
        expected_resolution_levels=expected_resolution_levels,
        expected_current_resolution_level=expected_current_resolution_level,
        expected_shape=expected_shape,
        expected_dtype=expected_dtype,
        expected_dims_order=expected_dims_order,
        expected_channel_names=expected_channel_names,
        expected_physical_pixel_sizes=expected_physical_pixel_sizes,
        expected_metadata_type=expected_metadata_type,
    )

    # Check for file pointers
    check_local_file_not_open(image_container)

    return image_container


def run_multi_scene_image_read_checks(
    ImageContainer: Type[ImageContainer],
    image: PathLike,
    first_scene_id: Union[str, int],
    first_scene_shape: Tuple[int, ...],
    first_scene_dtype: np.dtype,
    second_scene_id: Union[str, int],
    second_scene_shape: Tuple[int, ...],
    second_scene_dtype: np.dtype,
    allow_same_scene_data: bool = True,
) -> ImageContainer:
    """
    A suite of tests to ensure that data is reset when switching scenes.
    """
    # Read file
    image_container = ImageContainer(image, fs_kwargs=dict(anon=True))

    check_local_file_not_open(image_container)
    check_can_serialize_image_container(image_container)

    # Set scene
    image_container.set_scene(first_scene_id)

    # Check basics
    if isinstance(first_scene_id, str):
        assert image_container.current_scene == first_scene_id
    else:
        assert image_container.current_scene_index == first_scene_id
    assert image_container.shape == first_scene_shape
    assert image_container.dtype == first_scene_dtype

    # Check that the shape and dtype are expected after reading in full
    first_scene_data = image_container.data
    assert first_scene_data.shape == first_scene_shape
    assert first_scene_data.dtype == first_scene_dtype

    check_local_file_not_open(image_container)
    check_can_serialize_image_container(image_container)

    # Change scene
    image_container.set_scene(second_scene_id)

    # Check basics
    if isinstance(second_scene_id, str):
        assert image_container.current_scene == second_scene_id
    else:
        assert image_container.current_scene_index == second_scene_id
    assert image_container.shape == second_scene_shape
    assert image_container.dtype == second_scene_dtype

    # Check that the shape and dtype are expected after reading in full
    second_scene_data = image_container.data
    assert second_scene_data.shape == second_scene_shape
    assert second_scene_data.dtype == second_scene_dtype

    # Check that the first and second scene are not the same
    if not allow_same_scene_data:
        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_equal,
            first_scene_data,
            second_scene_data,
        )

    check_local_file_not_open(image_container)
    check_can_serialize_image_container(image_container)

    return image_container


def run_no_scene_name_image_read_checks(
    ImageContainer: Type[ImageContainer],
    image: PathLike,
    first_scene_id: Union[str, int],
    first_scene_dtype: np.dtype,
    second_scene_id: Union[str, int],
    second_scene_dtype: np.dtype,
    allow_same_scene_data: bool = True,
) -> ImageContainer:
    """
    A suite of tests to check that scene names are auto-filled when not present, and
    scene switching is reflected in current_scene_index.
    """
    # Read file
    image_container = ImageContainer(image, fs_kwargs=dict(anon=True))

    check_local_file_not_open(image_container)
    check_can_serialize_image_container(image_container)

    # Set scene
    image_container.set_scene(0)

    assert image_container.current_scene_index == 0

    # Check basics
    if isinstance(first_scene_id, str):
        assert image_container.current_scene == first_scene_id
    else:
        assert image_container.current_scene_index == first_scene_id
    assert image_container.dtype == first_scene_dtype

    # Check that the shape and dtype are expected after reading in full
    first_scene_data = image_container.data
    assert first_scene_data.dtype == first_scene_dtype

    check_local_file_not_open(image_container)
    check_can_serialize_image_container(image_container)

    # Change scene
    image_container.set_scene(1)

    assert image_container.current_scene_index == 1

    # Check basics
    if isinstance(second_scene_id, str):
        assert image_container.current_scene == second_scene_id
    else:
        assert image_container.current_scene_index == second_scene_id
    assert image_container.dtype == second_scene_dtype

    # Check that the shape and dtype are expected after reading in full
    second_scene_data = image_container.data
    assert second_scene_data.dtype == second_scene_dtype

    # Check that the first and second scene are not the same
    if not allow_same_scene_data:
        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_equal,
            first_scene_data,
            second_scene_data,
        )

    check_local_file_not_open(image_container)
    check_can_serialize_image_container(image_container)

    return image_container



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import List

from .reader import Reader

###############################################################################


class ReaderMetadata(ABC):
    @staticmethod
    @abstractmethod
    def get_supported_extensions() -> List[str]:
        pass

    @staticmethod
    @abstractmethod
    def get_reader() -> Reader:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import List, NamedTuple, Optional, Union

import dask.array as da
import numpy as np
import xarray as xr

###############################################################################

# IO Types
PathLike = Union[str, Path]
ArrayLike = Union[np.ndarray, da.Array]
MetaArrayLike = Union[ArrayLike, xr.DataArray]
ImageLike = Union[
    PathLike, ArrayLike, MetaArrayLike, List[MetaArrayLike], List[PathLike]
]


# Image Utility Types
class PhysicalPixelSizes(NamedTuple):
    Z: Optional[float]
    Y: Optional[float]
    X: Optional[float]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional


class ConflictingArgumentsError(Exception):
    """
    This exception is returned when 2 arguments to the same function are in conflict.
    """


class InvalidDimensionOrderingError(Exception):
    """
    A general exception that can be thrown when handling dimension ordering or
    validation. Should be provided with a message for the user to be given more context.
    """


class UnexpectedShapeError(Exception):
    """
    A general exception that can be thrown when handling shape validation.
    Should be provided with a message for the user to be given more context.
    """


class UnsupportedFileFormatError(Exception):
    """
    This exception is intended to communicate that the file extension is not one of
    the supported file types and cannot be parsed with AICSImage.
    """

    def __init__(self, reader_name: str, path: str, msg_extra: Optional[str] = None):
        super().__init__()
        self.reader_name = reader_name
        self.path = path
        self.msg_extra = msg_extra

    def __str__(self) -> str:
        msg = f"{self.reader_name} does not support the image: '{self.path}'."

        if self.msg_extra is not None:
            msg = f"{msg} {self.msg_extra}"

        return msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Any, Callable, List, Literal, Mapping, Optional, Tuple, Union

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from bioio_base import exceptions, noop_reader, transforms, types


@pytest.mark.parametrize("array_maker", [np.zeros, da.zeros])
@pytest.mark.parametrize(
    "data_shape, given_dims, return_dims, other_args, expected_shape",
    [
        (
            (10, 1, 5, 6, 200, 400),
            "STCZYX",
            "CSZYX",
            {},
            (5, 10, 6, 200, 400),
        ),
        ((6, 200, 400), "ZYX", "STCZYX", {}, (1, 1, 1, 6, 200, 400)),
        ((6, 200, 400), "ZYX", "ZCYSXT", {}, (6, 1, 200, 1, 400, 1)),
        ((6, 200, 400), "ZYX", "CYSXT", {"Z": 2}, (1, 200, 1, 400, 1)),
        (
            (6, 200, 400),
            "ZYX",
            "ZCYSXT",
            {"Z": [0, 1]},
            (2, 1, 200, 1, 400, 1),
        ),
        (
            (6, 200, 400),
            "ZYX",
            "ZCYSXT",
            {"Z": (0, 1)},
            (2, 1, 200, 1, 400, 1),
        ),
        (
            (6, 200, 400),
            "ZYX",
            "ZCYSXT",
            {"Z": range(2)},
            (2, 1, 200, 1, 400, 1),
        ),
        (
            (6, 200, 400),
            "ZYX",
            "ZCYSXT",
            {"Z": slice(0, 2, 1)},
            (2, 1, 200, 1, 400, 1),
        ),
        (
            (6, 200, 200, 3),
            "ZYXS",
            "YX",
            {"Z": 0, "S": 0},
            (200, 200),
        ),
        ((2, 2, 2), "ABI", "ZCYSXT", {}, (1, 1, 1, 1, 1, 1)),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXC",
            {"Z": 7},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZ",
            {"Z": 7},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": 7},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCX",
            {"Z": [0, 1, 4]},
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": [0, 1, 7]},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": (0, 1, 7)},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": range(7)},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": slice(0, 7, 2)},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": [0, 1, -7]},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": (0, 1, -7)},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": range(0, -8, -1)},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
        pytest.param(
            (6, 200, 400),
            "ZYX",
            "TYXCZX",
            {"Z": slice(-7, 0, 2)},
            None,
            marks=pytest.mark.xfail(raises=IndexError),
        ),
    ],
)
def test_reshape_data_shape(
    array_maker: Callable,
    data_shape: Tuple[int, ...],
    given_dims: str,
    return_dims: str,
    other_args: Any,
    expected_shape: Tuple[int, ...],
) -> None:
    data = array_maker(data_shape)

    actual = transforms.reshape_data(
        data=data, given_dims=given_dims, return_dims=return_dims, **other_args
    )
    assert actual.shape == expected_shape

    # Check that the output data is the same type as the input
    assert type(actual) == type(data)


@pytest.mark.parametrize(
    "data, given_dims, return_dims, idx_in, idx_out",
    [
        (
            np.random.rand(10, 1, 5, 6, 200, 400),
            "STCZYX",
            "CSZYX",
            (5, 0, 3, 3, ...),
            (3, 5, 3, ...),
        ),
        (
            da.random.random((10, 1, 5, 6, 200, 400)),
            "STCZYX",
            "CSZYX",
            (5, 0, 3, 3, ...),
            (3, 5, 3, ...),
        ),
        (
            np.zeros((6, 200, 400)),
            "ZYX",
            "STCZYX",
            (..., 100, 200),
            (0, 0, 0, ..., 100, 200),
        ),
        (
            da.zeros((6, 200, 400)),
            "ZYX",
            "STCZYX",
            (..., 100, 200),
            (0, 0, 0, ..., 100, 200),
        ),
        (
            np.zeros((6, 200, 400)),
            "ZYX",
            "ZCYSXT",
            (3, 100, ...),
            (3, 0, 100, 0, ..., 0),
        ),
        (
            da.zeros((6, 200, 400)),
            "ZYX",
            "ZCYSXT",
            (3, 100, ...),
            (3, 0, 100, 0, ..., 0),
        ),
    ],
)
def test_reshape_data_values(
    data: types.ArrayLike,
    given_dims: str,
    return_dims: str,
    idx_in: Tuple[int, ...],
    idx_out: Tuple[int, ...],
) -> None:
    slice_in = data[idx_in]
    actual = transforms.reshape_data(
        data=data, given_dims=given_dims, return_dims=return_dims
    )

    # Handle dask vs numpy
    if isinstance(actual, da.core.Array):
        slice_in_computed = slice_in.compute()
        actual_computed = actual.compute()
    else:
        slice_in_computed = slice_in
        actual_computed = actual

    np.testing.assert_array_equal(slice_in_computed, actual_computed[idx_out])

    # Check that the output data is the same type as the input
    assert type(actual) == type(slice_in)


# Arrays used for value checking on kwarg provided reshape_data
NP_ONES = np.ones((10, 10))
TEST_NDARRAY = np.stack([NP_ONES * i for i in range(7)])
DA_ONES = da.ones((10, 10))
TEST_DARRAY = da.stack([DA_ONES * i for i in range(7)])


@pytest.mark.parametrize("data", [TEST_NDARRAY, TEST_DARRAY])
@pytest.mark.parametrize(
    "given_dims, return_dims, other_args, getitem_ops_for_expected, transposer",
    [
        # Just dimension selection
        ("ZYX", "YX", {}, 0, None),
        ("ZYX", "YX", {"Z": 1}, 1, None),
        ("ZYX", "ZYX", {"Z": [0, 1]}, [0, 1], None),
        ("ZYX", "ZYX", {"Z": (0, 1)}, [0, 1], None),
        ("ZYX", "ZYX", {"Z": [0, -1]}, [0, -1], None),
        ("ZYX", "ZYX", {"Z": (0, -1)}, [0, -1], None),
        ("ZYX", "ZYX", {"Z": range(2)}, [0, 1], None),
        ("ZYX", "ZYX", {"Z": range(0, 6, 2)}, [0, 2, 4], None),
        ("ZYX", "ZYX", {"Z": slice(0, 6, 2)}, [0, 2, 4], None),
        ("ZYX", "ZYX", {"Z": slice(6, 3, -1)}, [6, 5, 4], None),
        ("ZYX", "ZYX", {"Z": slice(-1, 3, -1)}, [6, 5, 4], None),
        # Dimension selection and order swap
        (
            "ZYX",
            "YXZ",
            {"Z": (0, -1)},
            [0, -1],
            (1, 2, 0),
        ),
        (
            "ZYX",
            "YXZ",
            {"Z": range(0, 6, 2)},
            [0, 2, 4],
            (1, 2, 0),
        ),
    ],
)
def test_reshape_data_kwargs_values(
    data: types.ArrayLike,
    given_dims: str,
    return_dims: str,
    other_args: Mapping[str, Union[int, List[int], range, slice]],
    getitem_ops_for_expected: List[int],
    transposer: Tuple[int],
) -> None:
    actual = transforms.reshape_data(
        data=data,
        given_dims=given_dims,
        return_dims=return_dims,
        **other_args,
    )

    expected = data[getitem_ops_for_expected]

    if transposer is not None:
        if isinstance(data, np.ndarray):
            expected = np.transpose(expected, transposer)
        else:
            expected = da.transpose(expected, transposer)

    # Check that the output data is the same type as the input
    assert type(actual) == type(expected)

    if isinstance(actual, da.core.Array):
        actual = actual.compute()
        expected = expected.compute()

    # Check actual data
    np.testing.assert_array_equal(actual, expected)


@pytest.mark.parametrize(
    "data, given_dims, return_dims, expected_shape",
    [
        (np.zeros((1, 2, 3, 4, 5, 6)), "STCZYX", "XYZCTS", (6, 5, 4, 3, 2, 1)),
        (da.zeros((1, 2, 3, 4, 5, 6)), "STCZYX", "XYZCTS", (6, 5, 4, 3, 2, 1)),
        (np.zeros((1, 2, 3)), "ZYX", "ZXY", (1, 3, 2)),
        (da.zeros((1, 2, 3)), "ZYX", "ZXY", (1, 3, 2)),
        pytest.param(
            np.zeros((6, 200, 400)),
            "ZYX",
            "TYXC",
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        pytest.param(
            da.zeros((6, 200, 400)),
            "ZYX",
            "TYXC",
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        pytest.param(
            np.zeros((6, 200, 400)),
            "ZYX",
            "TYXCZ",
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
        pytest.param(
            da.zeros((6, 200, 400)),
            "ZYX",
            "TYXCZ",
            None,
            marks=pytest.mark.xfail(raises=exceptions.ConflictingArgumentsError),
        ),
    ],
)
def test_transpose_to_dims(
    data: types.ArrayLike,
    given_dims: str,
    return_dims: str,
    expected_shape: Tuple[int, ...],
) -> None:
    actual = transforms.transpose_to_dims(
        data=data, given_dims=given_dims, return_dims=return_dims
    )
    assert actual.shape == expected_shape

    # Check that the output data is the same type as the input
    assert type(actual) == type(data)


def get_data_reference(
    mode: str,
    mismatch: str,
    scene_character: str,
    scene_coord_values: str,
    select_scenes: Optional[tuple[Union[int, str], ...]] = None,
) -> tuple[list[types.MetaArrayLike], types.MetaArrayLike]:
    # set up the core values for the dataset
    shape = (3, 4, 5, 6, 7, 8)
    scene_name_to_idx = dict(
        zip([f"Image:{i}" for i in range(shape[0])], range(shape[0]))
    )
    data = np.arange(np.prod(shape), dtype="uint16").reshape(shape)

    # convert to dask if necessary
    if "dask" in mode:
        data = da.from_array(data)

    # split into list for image container construction
    data = [_ for _ in data]

    # mess with a scene for tests
    if mismatch == "shape":
        data[1] = data[1][:-1]
    elif mismatch == "dtype":
        data[1] = data[1].astype(int)

    # select scenes from argumnet or to make valid call to stack
    if select_scenes is None:
        reference = (
            np.stack(data) if mismatch == "none" else np.stack([data[0], data[2]])
        )
    else:
        if isinstance(select_scenes[0], int):
            reference = np.stack([data[i] for i in select_scenes])
        elif isinstance(select_scenes[0], str):
            inds = (scene_name_to_idx[str(s)] for s in select_scenes)
            reference = np.stack([data[i] for i in inds])
    # Assign coords and dims for xarray types
    if "xarray" in mode:
        data = [xr.DataArray(x, dims=list("TCZYX")) for x in data]
        reference = xr.DataArray(reference, dims=list((scene_character, *"TCZYX")))

    return data, reference


@pytest.mark.parametrize(
    "mode, scene_character, scene_coord_values",
    [
        ("data", "I", "index"),
        ("data", "U", "names"),
        ("dask_data", "I", "index"),
        ("xarray_data", "I", "index"),
        ("xarray_data", "U", "index"),
        pytest.param(
            "xarray_data", "T", "index", marks=pytest.mark.xfail(raises=ValueError)
        ),
        ("xarray_dask_data", "I", "index"),
        ("xarray_dask_data", "U", "index"),
        pytest.param(
            "xarray_dask_data",
            "T",
            "index",
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_generate_stack_stacking(
    mode: Literal["data", "dask_data", "xarray_data", "xarray_dask_data"],
    scene_character: str,
    scene_coord_values: str,
) -> None:
    data, reference = get_data_reference(
        mode, "none", scene_character, scene_coord_values
    )
    container = noop_reader.NoopReader(data)
    stack = transforms.generate_stack(
        container, mode, False, None, scene_character, scene_coord_values
    )
    if "xarray" in mode:
        xr.testing.assert_allclose(stack, reference)
    else:
        np.testing.assert_allclose(stack, reference)


@pytest.mark.parametrize(
    "mode", ["data", "xarray_data", "dask_data", "xarray_dask_data"]
)
@pytest.mark.parametrize(
    "mismatch, drop_non_matching_scenes, select_scenes",
    [
        ("none", True, None),
        ("none", False, (0, 2)),
        ("shape", False, (0, 2)),
        ("dtype", False, (0, 2)),
        ("none", False, tuple(f"Image:{i}" for i in [0, 2])),
        ("shape", False, tuple(f"Image:{i}" for i in [0, 2])),
        ("dtype", False, tuple(f"Image:{i}" for i in [0, 2])),
    ],
)
def test_generate_stack_mismatch_and_drop(
    # Test dropping and catching mismatched images
    mode: Literal["data", "dask_data", "xarray_data", "xarray_dask_data"],
    mismatch: str,
    drop_non_matching_scenes: bool,
    select_scenes: Optional[tuple[Union[int, str], ...]],
) -> None:
    data, reference = get_data_reference(mode, mismatch, "I", "index", select_scenes)
    container = noop_reader.NoopReader(data)

    stack = transforms.generate_stack(
        container, mode, drop_non_matching_scenes, select_scenes, "I", "index"
    )
    if "xarray" in mode:
        xr.testing.assert_allclose(stack, reference)
    else:
        np.testing.assert_allclose(stack, reference)


@pytest.mark.parametrize(
    "list_to_test, expected",
    [
        ([0], slice(0, 1, None)),
        ([4], slice(4, 5, None)),
        ([0, 1], slice(0, 2, 1)),
        ([1, 0], [1, 0]),
        ([3, 5], slice(3, 6, 2)),
        ([8, 9, 11], [8, 9, 11]),
        ([15, 20, 25], slice(15, 26, 5)),
        ((0,), slice(0, 1, None)),
        ((0, 1), slice(0, 2, 1)),
        ((1, 0), (1, 0)),
        ((3, 5), slice(3, 6, 2)),
        ((8, 9, 11), (8, 9, 11)),
        ((15, 20, 25), slice(15, 26, 5)),
    ],
)
def test_convert_list_to_slice(
    list_to_test: Union[List, Tuple], expected: Union[int, List, slice, Tuple]
) -> None:
    assert transforms.reduce_to_slice(list_to_test) == expected
